﻿#!/usr/bin/perl -w
use strict;
use warnings;
use utf8;
use open qw(:std :utf8);
use integer;

use Getopt::Long;
use Pod::Usage;


# Filter_2_03_find_forbidden_characters.pl
# 2010-08-14
# Wolfgang Schmidle

# This script checks whether the text contains characters that will make 
# the workflow scripts either break oder produce unexpected results

# input: raw text, normally as a whole
# output: a (possibly empty) list of warnings

my $name = "Filter_2_03_find_forbidden_characters";
my $textfilterParameters = "Filter_parameters/$name.txt";


# get command line options

my $dir = ""; 
my @forbidden;

my $man = 0;
my $help = 0;
my $parameterformat = 'dir=s,$'."\n".
			'forbidden=s,@,forbidden'."\n";

sub readParameters {

	GetOptions("dir=s" => \$dir,
			   "forbidden=s" => \@forbidden,
			   "parameterformat" => sub { print $parameterformat; exit; },
			   'help|?' => \$help, man => \$man) or pod2usage(2);
	
	pod2usage(1) if $help;
	pod2usage(-exitstatus => 0, -verbose => 2) if $man;
	
	@forbidden = split(/,/,join(',',@forbidden));
	
	if ($dir) {
		unless ($dir =~ m!/$!) { $dir .= "/"; }
	}
}

sub isTextfilter {
	return (exists $ENV{'BB_DOC_NAME'})
}

readParameters();

# ohne Dateiname: brich mit USAGE ab. 
unless ($#ARGV > -1) { pod2usage(1); } 


# text input

my @text;
while(<>) { push @text, $_; }


# read the "parameters" block

my $inText = 0;
my $inParameters = 0;

foreach (@text) {
	last if m!<pb[ ^]!;
	
	if (m!parameters:!) { $inParameters = 1; }
	if ($inParameters) {
		last if m!^ *$!;
		while (m!forbidden *= *([^ \n]+)!) { push @forbidden, $1; }
	}
}

# read the textfilter parameters

if (isTextfilter()) {

	open (PARA, $textfilterParameters) or die "2.03, used as textfilter: can't find the parameter list!\n";
	@ARGV = <PARA>;
	close(PARA);
	chomp @ARGV;

	readParameters();
}


# read helper files

unless (@forbidden) { die "2.03: no list of forbidden characters supplied!\n"; }

if ($dir) { 
	foreach (@forbidden) { $_ = $dir . $_; }
}
@ARGV = @forbidden;

my $reserved = "";
my %charSuggestions = ();
my @blockBlacklist = ();
my @charBlacklist = ();

while (<>) {
	chomp;
	s!#.*$!!; # remove comments
	s! +$!!;
	next if m!^$!;

	if (m!reserved: *([^ ]+)!) { $reserved .= $1; }
	if (m!charSuggestion: *([^ ]+) *= *([^ ]+)!) { $charSuggestions{$1} = $2; }
	if (m!blockBlacklist: *([^ ]+) *= *([^ ]+)!) { push @blockBlacklist, $1; }
	if (m!charBlacklist: *([^ ]+) *= *([^ ]+)!) { push @charBlacklist, $1; }
}


# go through the text

my $i = 0;
$inText = 0;

LINE: foreach (@text) {
	$i++;
	
	$inText = 1 if m!<pb[ >]!;
	next unless $inText;
	
	if ($reserved) { 
		if (m!([$reserved])!) { 
			warn "Line $i contains the reserved character \"$1\": \n$_\n"; 
			next LINE;
		}
	}
	
	foreach my $char (keys %charSuggestions) {
		if (m!$char!)  { 
			warn "Line $i contains \"$char\", which should be replaced by \"$charSuggestions{$char}\":\n$_\n"; 
			next LINE;
		}
	}

	foreach my $block (@blockBlacklist) {
		if (m!([\p{In$block}])!)  { 
			warn "Line $i contains \"$1\" from Unicode block \"$block\":\n$_\n"; 
			next LINE;
		}
	}

	foreach my $char (@charBlacklist) {
		if (m!$char!)  { 
			warn "Line $i contains \"$char\":\n$_\n"; 
			next LINE;
		}
	}
}

die "That's all.\n";


# TO DO:

# weitere Kandidaten: 
# ► in GeometricShapes (in WO1, Archimedes 1565)

# insert_reg verwendet auch ¢ und £. Ebenfalls verbieten? Oder andere Zeichen suchen?


__END__

=head1 NAME

Filter_2_03_find_forbidden_characters

=head1 SYNOPSIS

perl Filter_2_03_find_forbidden_characters.pl [options] [file]

 Options:
   -help       brief help message
   -man        full documentation
   -dir        path to the additional files 
   -forbidden   adds a list of forbidden characters

=head1 OPTIONS
=over 8

=item B<-help>
Print a brief help message and exits.

=item B<-man>
Prints the manual page and exits.

=item B<-forbidden>
Adds a list of forbidden characters. The order of lists is important. 

=back

=head1 DESCRIPTION

B<Filter_2_03_find_forbidden_characters> will check the input file for forbidden characters.

For the parameter fromat in the helper files see ...

=head1 TO DO

...

=cut
