﻿#!/usr/bin/perl -w
use strict;
use warnings;
use utf8;
use open qw(:std :utf8);
use integer;
use Unicode::Normalize;

use Getopt::Long;
use Pod::Usage;


# Filter_5_01_insert_reg.pl
# 2010-08-11
# Wolfgang Schmidle

# beware: 
# Mac OS X 10.6 still includes Perl 5.10.0, which implements Unicode 5.0 
# (see https://it-dev.mpiwg-berlin.mpg.de/tracs/mpdl-project-content/wiki/mac-perl-unicode)

my $name = "Filter_5_01_insert_reg";

# the following variables may be changed

my $que1 = "(?:q;|)"; # alle Formen von -que ohne Akut
my $que2 = "(?:q́;|'que|́)"; # alle Formen von -que mit Akut (beachte: vor Variablen kann das ; fehlen)
# neu: -q́ue wird erst in der Normalisierung entfernt

# my %wordlistContext = (
# 		"tñ" => "tantum" # check for "tamen" first!
# 		);

my $klein = "a-zſæœèòäöüßó"; 
# $klein .= "ãẽĩõũꝰę"; # sollten Vokale mit Tilde hier überhaupt dabei sein? Das sind doch nur Zeichen, die nicht normalisiert werden?? Immerhin schaden sie wohl auch nicht?
my $gross = "A-ZÄÖÜÆŒ";
my $vor = " (>;:";
my $nach = " )<.,:;!?&\n";  # Vorsicht bei . und : 


# the following variables shouldn't be changed

my $vowelTilde = "[ãẽĩõũ]";
my $vowelTildeEndWithM = "[ãẽũ]";
my $vowelTildeEndWithN = "[õ]";   # resolve ĩ via the wordlist
my $withM = "[pbm]";
my $withN = "[sſtcdfginquve]";
my %noTilde = ( "ã" => "a", "ẽ" => "e", "ĩ" => "i", "õ" => "o", "ũ" => "u" );

	# Parameter für Alvarus: Makron bewirkt das gleiche wie Tilde
	# (in anderen Texten könnte man gleich Makron durch Tilde ersetzen!)
	$vowelTilde = "[ãẽĩõũāēīōū]";
	$vowelTildeEndWithM = "[ãẽũāēū]";
	$vowelTildeEndWithN = "[õō]";
	%noTilde = ( 
		"ã" => "a", "ẽ" => "e", "ĩ" => "i", "õ" => "o", "ũ" => "u",
		"ā" => "a", "ē" => "e", "ī" => "i", "ō" => "o", "ū" => "u"
	);

my $hyphen = "[-­]"; # normales hyphen und soft hyphen
my $lb = "<lb/>";
my $pb = "\n?<pb [^>]+>\n?";
my $linebreak = "$hyphen(?:$lb|$pb)";

my $kl = "(?:[$klein]|$linebreak)*";
my $gr = "[$gross]?";
my $vo = "(?<=[$vor])"; 
my $na = "(?=[$nach])";

my $origDef = "[^¤]+";
my $normDef = "[^¤]*";
my $faithfulDef = "[^¤]*";
my $typeDef = "[^¤]+";

my $reg = "¤B¤($origDef)¤($normDef)¤($faithfulDef)¤($typeDef)¤E¤";


# get command line options

my $dir = ""; 
my @wordlist; 
my @simple; 
my @context; # Kontextregeln werden noch nicht so übergeben, es gibt auch noch kein Datenformat für Kontextregeln

my $man = 0;
my $help = 0;
my $parameterformat = 'dir=s,$'."\n".
			'wordlist=s,@,reg-wordlist'."\n".
			'simple=s,@,reg-simple'."\n";

sub readParameters {

	GetOptions("dir=s" => \$dir,
			   "wordlist=s" => \@wordlist,
			   "simple=s" => \@simple,
			   "parameterformat" => sub { print $parameterformat; exit; },
			   'help|?' => \$help, man => \$man) or pod2usage(2);
	
	pod2usage(1) if $help;
	pod2usage(-exitstatus => 0, -verbose => 2) if $man;
	
	@wordlist = split(/,/,join(',',@wordlist));
	@simple = split(/,/,join(',',@simple));
	@context = split(/,/,join(',',@context));
	
	if ($dir) {
		unless ($dir =~ m!/$!) { $dir .= "/"; }
	}
}

sub anyParameterSet {
	return (@wordlist or @simple or @context)  # Kriterium gut genug?
}

readParameters();

# ohne Dateiname: brich mit USAGE ab. 
# offenbar auch Textfilter-tauglich. Testen: auch Pipe-tauglich?
unless ($#ARGV > -1) { pod2usage(1); } 

# Text einlesen

my @text;
my $isLongline = 0;
my $longline = "";

while(<>) { 
	if ($isLongline) {
		$longline .= $_;
		next if m!<anchor !;
		next if m!<pb !;
		next if m!<cb[ /]!;
		next if m![^>]\n!;
		
		push @text, $longline;
		$isLongline = 0;
		next;
	}

	if (m![^>]\n!) { 
		$isLongline = 1; 
		$longline = $_;
		next;
	}
	push @text, $_; 
}


# add parameters from <parameters>
# (after the text has been read)
# note that commandline parameter "wordlist" is "reg.wordlist" in <parameters>, etc.

my $inParameters = 0;

foreach (@text) {

	last if m!<text[ ^]!;
	last if m!</parameters>!;
	
	if (m!<parameters>!) { $inParameters = 1; }
	if ($inParameters) {
		while (m!reg\.wordlist *= *([^, <]+)[, <]!g) { push @wordlist, $dir.$1; }
		while (m!reg\.simple *= *([^, <]+)[, <]!g) { push @simple, $dir.$1; }
	}
}


# Parameter-Übergabe bei Verwendung als Textfilter:
# bei Verwendung als Texfilter gibt es keine Commandline-Parameter, und 
# <parameters> muss nicht in der Selection enthalten sein.

# unless (anyParameterSet()) {

if (exists $ENV{'BB_DOC_NAME'}) { # i.e. if run as textfilter from BBEdit

	open (PARA, "Filter_parameters/$name.txt") or die "Can't find the parameter list!\n";
	@ARGV = <PARA>;
	close(PARA);
	chomp @ARGV;

	readParameters();

# 	$dir = "/Users/wschmidle/Desktop/Schema/trunk/trunk/texts/aux";
# 
# 	@wordlist = (
# 		"reg-wordlist-lat.txt",
# 		"reg-wordlist-lat-hentisberi.txt",
# 		"reg-wordlist-lat-alvarus.txt"
# 	);
# 
# 	@simple = (
# 		"reg-simple-lat.txt",
# 		"reg-simple-lat-alvarus.txt"
# 	);

}


# Hilfsdateien einlesen

my %wordlistOtN = ();
my %wordlistFtO = ();
my %wordlistFtN = ();
my %wordlistAbbreviations = ();
my %regSimple = ();

if ($dir) { 
	foreach (@wordlist) { $_ = $dir . $_; }
}

@ARGV = @wordlist;

while (<>) {
	
	chomp;
	s!#.*$!!; # remove comments
	s! +$!!;
	next if m!^$!;
	
	if (m!^\[(.+?)\] *= *(.+)!) {
		$wordlistAbbreviations{$1} = $2;
		next;
	}
	if (m!(.+?) +\[(.+?)\] *= *(.*)!) {
		my $faithful = $1;
		my $orig = $2;
		my $norm = $3;
		$orig =~ s!\.!\\.!g;
		$faithful =~ s!\.!\\.!g;
		$wordlistFtO{$faithful} = $orig;
		$wordlistFtN{$faithful} = $norm;
		if ($faithful =~ m!^([a-zſ])!) {
			$wordlistFtO{ucfirst $faithful} = ucfirst $orig;
			$wordlistFtN{ucfirst $faithful} = ucfirst $norm;
		}
		next;
	}
	if (m!(.+?) *= *(.*)!) {
		my $orig = $1;
		my $norm = $2;
		while ($orig =~ m!\[(.+?)\]!g) {
			unless (exists $wordlistAbbreviations{$1}) { die "unknown abbreviation [$1]\n"; }
		}
		$orig =~ s!\[(.+?)\]!$wordlistAbbreviations{$1}!g;
		$orig =~ s!\.!\\.!g;
		$wordlistOtN{$orig} = $norm;
		if ($orig =~ m!^([a-zſ])!) {
			$wordlistOtN{ucfirst $orig} = ucfirst $norm;
		} 
	}
}


if ($dir) { 
	foreach (@simple) { $_ = $dir . $_; }
}

@ARGV = @simple;

while (<>) {
	
	chomp;
	s!#.*$!!; # remove comments
	s! +$!!;
	next if m!^$!;
	
	if (m!(.+?) *> *(.*)!) {
		my $orig = $1;
		my $norm = $2;
		$regSimple{$orig} = $norm;
	}
}



# Textzeilen bearbeiten

my $inText = 0;

my $original;

foreach (@text) {

	$original = $_;

# 	if (m!<text[ ^]!) { $inText = 1; }
# 	next unless $inText;
# man könnte unterscheiden: wenn ganze Datei, dann erst ab Text, wenn Ausschnitt, dann ganz
	
	# the following things are just a precaution
	$_ = NFC($_); 	
	if (m!([¤¥¢£])!) { die "Der Text enthält das reservierte Zeichen $1: $_\n"; }
	

	# Das allgemeine Format für Ersetzungen ist:  ¤B¤orig¤norm¤faithful¤type¤E¤
	# norm und faithful können auch leer sein!
	#
	# Beispiel: 
	# pręsẽs  --> 
	# pr¤ę¤ae¤simple¤sẽs --> 
	# pr¤ę¤ae¤simple¤s¤ẽ¤en¤context¤s --> 
	# <reg orig="pręsẽs" type="simple context">praesens</reg>  -->
	# <reg orig="pręsẽs" type="context">praesens</reg>  (falls alle "simple" entfernt werden)
	
	
	# beachte: ($key) und orig="$1", damit $que aufgelöst wird

# 	foreach my $key (keys %wordlistContext) {
# 		s!$vo($key)$na!<reg orig="$1" type="wordlist/context">$wordlistContext{$key}</reg>!g; 
# 	}

	# schütze bereits regularisierte Formen vor erneuter Regularisierung
	
	s!</reg>!£</reg>!g; # reicht für Wortlisten-Regularisierungen und -que bereits aus
	s!<reg !¢<reg !g; 
	$_ = "£$_¢";
	
	s! faithful="([^"]+)"! faithful="¢$1£"!g;
	
	
	foreach my $key (keys %wordlistFtO) {
		s!$vo($key)$na!¢<reg norm="$wordlistFtN{$key}" faithful="$1" type="wordlist">$wordlistFtO{$key}£</reg>!g; 
	}
	
	foreach my $key (keys %wordlistOtN) {
		s!$vo($key)$na!¢<reg norm="$wordlistOtN{$key}" type="wordlist">$1£</reg>!g; 
	}
	
	# -que: vor allen anderen simplen Ersetzungen
	s/($que2)(?![£¥])/¤B¤$1¤que¤$1¤simple¤E¤/g;
	# das gleiche mit que1: aber eigentlich nur, wenn ein Parameter gesetzt ist
	s/($que1)(?![£¥])/¤B¤$1¤que¤$1¤simple¤E¤/g;
	
	s!(q́)( <var>|\. | \.)!¤B¤$1¤que¤$1¤simple¤E¤$2!g; # Ausnahme: vor einer Zahl oder Variablen kann ; fehlen. 
	
	foreach my $key (keys %regSimple) {
		1 while s/(¢[^£]*$key)(?!¥)/$1¥/;
		s/($key)(?![¥\p{M}])/¤B¤$1¤$regSimple{$key}¤$1¤simple¤E¤/g;
		# \p{M} verhindert, dass ꝗ auch ꝗ̃ findet
	}

	# Tilde über Vokal	
	1 while s/(¢[^£]*$vowelTilde)(?!¥)/$1¥/;
	s!($vowelTilde)(?=(?:$linebreak)?¤$que2)!¤B¤$1¤$noTilde{$1}m¤$1¤context¤E¤!g; # vor -que: m
	s!($vowelTilde)(?=(?:$linebreak)?$withM)!¤B¤$1¤$noTilde{$1}m¤$1¤context¤E¤!g;
	s!($vowelTilde)(?=(?:$linebreak)?$withN)!¤B¤$1¤$noTilde{$1}n¤$1¤context¤E¤!g;
	s!($vowelTildeEndWithM)$na!¤B¤$1¤$noTilde{$1}m¤$1¤context¤E¤!g;
	s!($vowelTildeEndWithN)$na!¤B¤$1¤$noTilde{$1}n¤$1¤context¤E¤!g;
	# (rules for õr=orr and õl=oll if as soon as cases like this occur)
	
	# ↄ̨ = com/con
	1 while s/(¢[^£]*ↄ̨)(?!¥)/$1¥/;
	s!(ↄ̨)(?=(?:$linebreak)?$withM)!¤B¤$1¤com¤$1¤context¤E¤!g;
	s!(ↄ̨)(?=(?:$linebreak)?$withN)!¤B¤$1¤con¤$1¤context¤E¤!g;
	

	s![¥¢£]!!g;
	
	
	# Wörter mit vier Ersetzungen
	s!$vo($gr$kl)$reg($kl)$reg($kl)$reg($kl)$reg($kl)$na!<reg norm="$1$3$6$8$11$13$16$18$21" faithful="$1$4$6$9$11$14$16$19$21" type="$5 $10 $15 $20">$1$2$6$7$11$12$16$17$21</reg>!og;
	s!( norm="[^"]*$hyphen)(?:$lb|$pb)([^"]*")!$1 $2!g;

	# Wörter mit drei Ersetzungen
	s!$vo($gr$kl)$reg($kl)$reg($kl)$reg($kl)$na!<reg norm="$1$3$6$8$11$13$16" faithful="$1$4$6$9$11$14$16" type="$5 $10 $15">$1$2$6$7$11$12$16</reg>!og;
	s!( norm="[^"]*$hyphen)(?:$lb|$pb)([^"]*")!$1 $2!g;

	# Wörter mit zwei Ersetzungen
	s!$vo($gr$kl)$reg($kl)$reg($kl)$na!<reg norm="$1$3$6$8$11" faithful="$1$4$6$9$11" type="$5 $10">$1$2$6$7$11</reg>!og;
	s!( norm="[^"]*$hyphen)(?:$lb|$pb)([^"]*")!$1 $2!g;

	# Wörter mit einer Ersetzung
	s!$vo($gr$kl)$reg($kl)$na!<reg norm="$1$3$6" faithful="$1$4$6" type="$5">$1$2$6</reg>!og;
	s!( norm="[^"]*$hyphen)(?:$lb|$pb)([^"]*")!$1 $2!g;
	
#	if (m!¤!) { die "Fehler beim Einfügen von <reg>: $_\n"; }

	# entferne überflüssige @norm und @faithful
	# TODO: Zeilenumbrüche im Wort
	1 while (s!<reg norm="([^"]*)" faithful="([^"]*)" type="([^"]*)">\1</reg>!<reg faithful="$2" type="$3">$1</reg>!);
	1 while (s!<reg norm="([^"]*)" faithful="([^"]*)" type="([^"]*)">\2</reg>!<reg norm="$1" type="$3">$2</reg>!);


	# verschiebe PUA-Zeichen in das @faithful-Attribut: 
	# Beispiel: 
	# vorher: <reg norm="protectorique" type="simple">protectori</reg>
	# nachher: <reg norm="protectorique" faithful="protectori" type="simple">protectoriqꝫ</reg>
	# Quiſs 
	# vtra-<lb/>
	
	s!<reg (norm="[^"]+") (type="[^"]+")>($gr$kl)́</reg>!<reg $1 faithful="$3́" $2>$3q́ꝫ</reg>!g;
	s!<reg (norm="[^"]+") (type="[^"]+")>($gr$kl)</reg>!<reg $1 faithful="$3" $2>$3qꝫ</reg>!g;
	s!(faithful="[^"]+)<lb/>([^"]+)!$1 $2!g;

#	1 while s!<reg (norm="[^"]+") (type="[^"]+")>($gr$kl)</reg>!<reg $1 faithful="$3" $2>$3qꝫ</reg>!;

	if (m!¤!) { 
		$_ = $original;
		unless (m/^!!! /) {
			$_ = "!!! " . $_;
		}
	}
}


# geänderten Text ausgeben

print @text;


# TO DO:
# tamen versus tantum: Besser getrennt in der Nachbearbeitung?
# pb ?
# ziehe den Abkürzungspunkt heraus, wenn es zugleich der Satzpunkt ist: Beispiel "aīa."

# ↄ̨ſtantium : besser simple statt wordlist?

# ↄ̨-<lb/>ſtantium: wird zwar mit <reg> markiert, aber der <lb/> verschwindet im regularisierten Text
# --> nee, stimmt gar nicht, sondern es wird nicht markiert
# --> kann man das irgendwie lösen? Anders gesagt: kann man eine Zuordnung der Wortstücke
# machen? Bei ↄ̨ſtantium geht das sicher, aber wie ist es bei merkwürdigen Abkürzungen?
# Kann man da dann wiederum erwarten, dass es keine <lb/> gibt?

# poem: anders markieren

# schau nach, ob alle Sachen in den Benedetti-Texten übernommen wurden! 
# Problem: <lb/> in einem Wort der wordlist
# Problem: Großbuchstabe in einem Wort der wordlist, z.B. "Qũo"

# unklare Wörter: bei der automatischen Verarbeitung besser als <reg orig="tñ" type="wordlist/context">tamen/tantum</reg> ?

# <reg> muss auch funktionieren, wenn xml:space schon eingefügt wurde; dürfte aber kein Problem sein

# muss auch mit &amp; statt ' umgehen können.

# Parameter: 
# - Alvarus: kein Akut bei -que, Benedetti: Akut, wenn es "und" meint, also nciht be denique
# - man kann Zeichen ergänzen, die als Variablen verwendet werden, zum Beispiel æ in Benedetti
# hierarchische Parameter, zum Beispiel "16. Jahrhundert", die dann Standardwerte für untergeorndete
#  Parameter bewirken (was wiederum durch explizite Setzung überschrieben werden kann)

# zum Alvarus-Hack: wenn es <parameters> nicht gibt, sollte das Skript nach einer Datei in einem festen 
# Verzeichnis suchen, zum Beispiel aux/parameters.txt

# und simple mit -->, sodass man alles auf einmal einlesen kann

# simple könnte schon auf die Wortliste wirken. nicht direkt, aber ungefähr. Oder: Parameter "makron gleich 
# tilde" könnte direkt auf die Wortliste wirken

# Hilfsdatei wie "Bischoff-paleography.txt" ?

# Wort isolieren, dann Regeln drauf anwenden.

# <reg norm="eamque" resp="eam" type="simple">eamqꝫ</reg> 
# ergibt eine Fehlermeldung, wenn man das Skript nochmal anwendet
# (liegt wohl am @resp, aber wie genau?)

# sollte erkennbar sein, aus welcher Datei die Regularisierungsregel kommt?

# die faithful-Formen in die Listen statt in im Skript erzeugen?
# die faithful-Formen gleich in die ¤ ¤¤ ¤¤¤-Schreibweise bringen?

# mache £ ¢ ¥ leichter verständlich

#  caducaꝫ 

# umsteigen von alle Wörter der Wortliste durchgehen zu alle Wörter des Textes durchgehen
# gehe die Zeile nur ein einziges Mal durch:
# gehe wortweise vor: finde das nächste Wort oder die beiden Teile des Wortes. Wende Regeln an.
# Überspringe bereits vorhandene <reg>
# auch bei <s>-Skript?



__END__

=head1 NAME

Filter_insert-reg - inserts <reg> into XML texts

=head1 SYNOPSIS

perl Filter_insert-reg [options] [file]

 Options:
   -help       brief help message
   -man        full documentation
   -dir        path to the addtional files 
   -wordlist   adds a wordlist
   -simple     adds a list of simple regularizations
   -context    adds a list of context-dependent regularizations

=head1 OPTIONS

=over 8

=item B<-help>

Print a brief help message and exits.

=item B<-man>

Prints the manual page and exits.

=item B<-wordlist>

Adds a wordlist. The order of wordlists is important. Wordlist rules are processed before simple and context rules.

=item B<-simple>

Adds a list of simple regularizations

=item B<-context>

Adds a list of context-dependent regularizations.

=back

=head1 DESCRIPTION

B<Filter_insert-reg> will add <reg> in the input XML file according to the rules in the script itself and the rules inthe additional files. 

The input file must be UTF-8.

=head1 TO DO

...

=cut
